/*
 * Copyright (C) 2000-2005 Serge Vakulenko, <vak@cronyx.ru>
 *
 * This file is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.
 *
 * You can redistribute this file and/or modify it under the terms of the GNU
 * General Public License (GPL) as published by the Free Software Foundation;
 * either version 2 of the License, or (at your discretion) any later version.
 * See the accompanying file "COPYING.txt" for more details.
 *
 * As a special exception to the GPL, permission is granted for additional
 * uses of the text contained in this file.  See the accompanying file
 * "COPY-UOS.txt" for details.
 */
#include <runtime/lib.h>
#include <kernel/uos.h>
#include <kernel/internal.h>

/*
 * Send the signal to the lock. All tasks waiting for the signal
 * on the lock are unblocked, possibly causing task switch.
 */
void
lock_signal (lock_t *m, void *message)
{
	int x;

	MACHDEP_INTR_DISABLE (&x);
	assert (STACK_GUARD (task_current));
	if (! m->waiters.head && ! m->groups.head) {
		MACHDEP_INTR_RESTORE (x);
		return;
	}
	lock_activate (m, message);
	MACHDEP_TASK_SWITCH ();
	MACHDEP_INTR_RESTORE (x);
}

/*
 * Wait for the signal on the lock. The calling task is blocked.
 * In the case the lock has associated IRQ number,
 * it is unmasked (permitted to happen).
 */
void *
lock_wait (lock_t *m)
{
	int x;

	MACHDEP_INTR_DISABLE (&x);
	assert (STACK_GUARD (task_current));
	assert (task_current->wait == 0);

	/* On pending irq, we must call fast handler. */
	if (m->irq) {
		if (m->irq->pending) {
			m->irq->pending = 0;

			if ((m->irq->handler) (m->irq->arg) == 0) {
				/* Unblock all tasks, waiting for irq. */
				lock_activate (m, 0);
				MACHDEP_INTR_RESTORE (x);
				return 0;
			}
		}
 		MACHDEP_INTR_ALLOW (m->irq->irq);
	}
	task_current->wait = m;
	task_dequeue (&task_active, task_current);
	task_enqueue (&m->waiters, task_current);

 	if (! m->master) {
		MACHDEP_TASK_SWITCH ();
 		MACHDEP_INTR_RESTORE (x);
 		return task_current->message;
 	}

 	/* The lock is hold by the current task - release it. */
 	assert (m->master == task_current);
 	assert (task_current->slaves.head == m);
	lock_dequeue (&m->master->slaves, m);

	/* Recalculate the value of task priority.
	 * It must be the maximum of base priority,
	 * and all slave lock priorities. */
	if (m->master->prio <= m->prio && m->master->base_prio < m->prio)
		task_recalculate_prio (m->master);

	m->master = 0;

	while (m->slaves.head) {
		task_t *t = m->slaves.head;
		assert (t->lock == m);
		t->lock = 0;
		task_dequeue (&m->slaves, t);
		task_enqueue (&task_active, t);
	}
	m->prio = 0;
	MACHDEP_TASK_SWITCH ();

	/* Acquire the lock again. */
	while (m->master) {
		/* Monitor is locked, block the task. */
		assert (task_current->lock == 0);
		task_current->lock = m;
		task_dequeue (&task_active, task_current);

		/* Put this task into the list of lock slaves. */
		task_enqueue (&m->slaves, task_current);

		/* Update the value of lock priority.
		 * It must be the maximum of all slave task priorities. */
		if (m->prio < task_current->prio) {
			m->prio = task_current->prio;

			/* Increase the priority of master task. */
			if (m->master->prio < m->prio)
				m->master->prio = m->prio;
		}

		MACHDEP_TASK_SWITCH ();
	}

	/* Put this lock into the list of task slaves. */
	m->master = task_current;
 	lock_enqueue (&task_current->slaves, m);

	/* Update the value of task priority.
	 * It must be the maximum of base priority,
	 * and all slave lock priorities. */
	if (task_current->prio < m->prio)
		task_current->prio = m->prio;

	MACHDEP_INTR_RESTORE (x);
	return task_current->message;
}
