/*
 * Copyright (C) 2000-2005 Serge Vakulenko, <vak@cronyx.ru>
 *
 * This file is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.
 *
 * You can redistribute this file and/or modify it under the terms of the GNU
 * General Public License (GPL) as published by the Free Software Foundation;
 * either version 2 of the License, or (at your discretion) any later version.
 * See the accompanying file "COPYING.txt" for more details.
 *
 * As a special exception to the GPL, permission is granted for additional
 * uses of the text contained in this file.  See the accompanying file
 * "COPY-UOS.txt" for details.
 */
#include <runtime/lib.h>
#include <kernel/uos.h>
#include <kernel/internal.h>

task_list_t task_active;			/* list of tasks ready to run */
task_t *task_current;				/* current running task */
task_t *task_idle;				/* background system task */
lock_irq_t lock_irq [MACHDEP_INTERRUPTS];	/* interrupt handlers */

static char task_idle_data [sizeof(task_t) + sizeof(long)];

/*
 * Utility functions.
 * Must be called with interrupts disabled.
 */
void
task_enqueue (task_list_t *list, task_t *elem)
{
	if (! list->head) {
		assert (list->tail == 0);
		elem->next = 0;
		elem->prev = 0;
		list->head = elem;
		list->tail = elem;
	} else {
		assert (list->tail != 0);
		elem->next = list->head;
		elem->prev = 0;
		list->head->prev = elem;
		list->head = elem;
	}
}

void
task_dequeue (task_list_t *list, task_t *elem)
{
	assert (list->head != 0);
	assert (list->tail != 0);
	if (list->head == elem) {
		list->head = elem->next;
		assert (elem->prev == 0);
	} else {
		assert (elem->prev != 0);
		elem->prev->next = elem->next;
	}
	if (list->tail == elem) {
		list->tail = elem->prev;
		assert (elem->next == 0);
	} else {
		assert (elem->next != 0);
		elem->next->prev = elem->prev;
	}
}

/*
 * Activate all waiters of the lock.
 */
void
lock_activate (lock_t *m, void *message)
{
	task_t *t;
	lock_slot_t *s;

	if (! m)
		return;
	while ((t = m->waiters.head)) {
		assert (t->wait == m);
		t->wait = 0;
		t->message = message;
		task_dequeue (&m->waiters, t);
		task_enqueue (&task_active, t);
	}
	/* Activate groups. */
	for (s=m->groups.head; s; s=s->next) {
		assert (s->lock == m);
		s->message = message;
		s->active = 1;
		t = s->group->waiter;
/*debug_printf ("lock_activate: slot %p msg %s task %s\n", */
/*s, message, t ? t->name : "<null>");*/
		if (t) {
			s->group->waiter = 0;
			task_enqueue (&task_active, t);
		}
	}
}

/*
 * Find the task with maximum priority,
 * and put it to the head of queue.
 * The core of the scheduler policy.
 */
task_t *
task_policy (void)
{
	int maxprio = 0;
	task_t *t, *r;

	assert (STACK_GUARD (task_current));
	r = task_idle;
	for (t=task_active.head; t; t=t->next) {
		if (t->prio >= maxprio) {
			r = t;
			maxprio = t->prio;
		}
	}
	task_current = r;
/*debug_printf ("task_policy: return `%s'\n", r->name);*/
	return r;
}

/*
 * Call user initialization routine uos_init(),
 * then create the idle task, and run the OS.
 * The idle task uses the default system stack.
 */
int
main (void)
{
	/* Create the idle task.
	 * Align stack on long word boundary. */
	task_idle = (task_t*) (((int) task_idle_data + sizeof (long) - 1) &
		~(sizeof (long) - 1));
	task_idle->stack[0] = STACK_MAGIC;
	task_idle->name = CONST("idle");
	task_current = task_idle;
	task_enqueue (&task_active, task_idle);

	/* Create user tasks. */
	uos_init ();

	/* Switch to the most priority task. */
	MACHDEP_TASK_SWITCH ();

	/* Idle task activity. */
#ifdef MACHDEP_IDLE
	MACHDEP_IDLE ();
#endif
}
